#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "bitar.h"
#include "rela.h"
#include "util.h"

struct rela
{
  int N;
  bitar_t *array;
};


/***********************************************************************
 *                                                                     *
 *                      PARTIAL ORDERING RELATION                      *
 *                                                                     *
 ***********************************************************************
 This file implements the partial ordering relation. We hold only half
 of the relation table since it's symmetrical with respect to the main
 diagonal. Also, R(i,i)= 0. We don't store the diagonal either.
 ***********************************************************************/


/***********************************************************************
 The table looks like the following. Xs represent which cells are stored.
     0 1 2 ..  N-1  Number of Xs  
   0   X X ..  X       N-1         = N-1 -0
   1     X ..  X       N-2         = N-1 -1
   2       ..  X       N-3         = N-1 -2
   .                   ..
 N-2           X        1          = N-1 -(N-2)

 So, for each row, the starting address is:
    (N-1)*r - (0+1+..r-1) =  (N-1)*r - (r-1)*r/2

 For a column c, the address of the cell within the row is:
    c - r -1 

 The total address is the sum of these two.

  r must be less than c.
 ***********************************************************************/

static int rela_address(rela_t *R,int r,int c)
{
  int N;
  N= R->N;
  return (N-1)*r - (r-1)*r/2 + c - r -1 ;
}

rela_t* rela_new(int N)
{
  int sz;
  rela_t *R;
  R= malloc(sizeof(*R));
  R->N= N;
  sz= rela_address(R,N-2,N-1)+1;
  R->array= bitar_new(sz);
  return R;
}

/***********************************************************************
  a must be less than b.
 ***********************************************************************/
static int rela_get(rela_t *R,int a,int b)
{
  return bitar_get(R->array,rela_address(R,a,b));
}

static void rela_set(rela_t *R, int a,int b)
{
  bitar_set(R->array, rela_address(R,a,b));
}

static void rela_clr(rela_t *R, int a,int b)
{
  bitar_clr(R->array, rela_address(R,a,b));
}

/***********************************************************************
 When asked for the comparison between element a and b:
 - if a<b,  we look at R(a,b). If this bit is 1, then line[a] < line[b].
   Otherwise these two are not related.
 - if a>b,  we look at R(b,a). If this bit is 1, then line[a] > line[b].
   Otherwise these two are not related.
 - if a==b, then these two are not in the relation, so 0 is returned. 
 ***********************************************************************/
int rela_compare(rela_t *R, int a, int b)
{
  if (a<b) { if (rela_get(R,a,b)) return -1; else return 0; }
  if (a>b) { if (rela_get(R,b,a)) return  1; else return 0; }
  return 0;
}

/***********************************************************************
 rela_greater(a,b) records the fact that  line[a] > line[b].

  line[x] cannot be greater than 
 ***********************************************************************/
void rela_greater(rela_t *R,int a,int b)
{
  if (a<=b) return ;
  rela_less(R, b,a); 
}

void rela_less(rela_t *R, int a, int b)
{
  if (a>=b) return ;
  rela_set(R,a,b);
}

rela_t *rela_dup(rela_t *R)
{
  rela_t *C;
  C= malloc(sizeof(*C));
  C->N= R->N;
  C->array= bitar_dup(R->array);
  return C;
}

void rela_remove(rela_t *R, int a,int b)
{
  if (a==b) return ;
  if (a<b) rela_clr(R,a,b);
     else  rela_clr(R,b,a);
}


void rela_apply_transitions(rela_t *R)
{
  int i,j,k,N,c;
  N= R->N;
  for(k=0;k<N;k++)
    for(i=0;i<N;i++)
      for(j=0;j<N;j++)
        if ((c=rela_compare(R,i,k)) == rela_compare(R,k,j))
        {
          if (c==1) rela_greater(R,i,j);
          else if (c==-1) rela_less(R,i,j);
        }
}

void rela_print(rela_t *R)
{
  int i,j,N;
  N= R->N;
  printf("  ");
  for(i=0;i<N;i++) printf("%d",i%10); printf("\n");
  for(i=0;i<N;i++)
  {
   printf("%d ", i%10);
   for(j=0;j<=i;j++) printf(" ");
   for(;j<N;j++)
     if (rela_get(R,i,j)) printf("X"); else printf(".");
   printf("\n");
  }
}

void rela_write(FILE *f,rela_t *R)
{
  int i,j,N;
  N= R->N;
  write_int(f, N);
  for(i=0;i<N-1;i++)
  {
    for(j=i+1;j<N;j++)
     if (rela_get(R,i,j)) fprintf(f,"1"); else fprintf(f,"0");
  }
}

rela_t* rela_read(FILE *f)
{
  uint32_t N,i,j;
  rela_t *R;
  char c;

  read_int(f,&N);
  R= rela_new(N);
  for(i=0;i<N-1;i++)
    for(j=i+1;j<N;j++)
      { fread(&c, 1,1, f); if (c=='1') rela_set(R,i,j); }
  return R;
}

int rela_count0(rela_t *R)
{
  return bitar_count0(R->array);
}

void rela_free(rela_t *R)
{
  bitar_free(R->array);
  free(R);
}
