#include <stdio.h>
#include <sys/select.h>
#include <unistd.h>
#include <sys/time.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdint.h>
#include "network.h"
#include "rela.h"
#include "util.h"
#include "relaser.h"
#include "tifa.h"

/***********************************************************************
 *                                                                     *
 *                        RELATION BASED SEARCH                        *
 *                                                                     *
 ***********************************************************************
 The state of the search is stored on disk when the program quits. This
 way I can run it sporadically. 

 In order to not search reordered networks more than once, I'll have to
 do some pruning. Networks A and B are reordered if they satisfy the 
 following conditions:
  - They have the same number of gates and levels
  - For each level in A, if you permute it, it becomes identical to the
    corresponding level in B.
 In short, I won't search the permutations of levels. I will simply pick
 one permutation and search from there. If I come across another
 permutation later on, I'll backtrack.

 Achieving this is quite easy. I just have to make sure that gates are
 inserted in a particular order into the level. If the new gate breaks
 this ordering, I won't insert it because another sequence which 
 satisfies the ordering will be eventually reached.

 The order I pick is the following: gates should be ordered by their
 upper line (numbered less). i.e. 3.4/5.6 obeys this ordering but
 5.6/3.4 doesn't. So, if I already have 5 in my upper_mask, then I can't
 continue with a gate such as 3.4 or 1.7.
 ***********************************************************************/

typedef struct { uint32_t a,b,c; } i3tuple_t;
typedef struct { uint32_t a,b; }   i2tuple_t;

typedef struct { uint32_t a,b;
            //---- Metrics ------//
                 int Mcom;    // complexity
                 int Minset;  // size of the transformed input set
                 int Mlevel;  // whether it creates a new level
               } candi_t;

/***********************************************************************
  Each senode contains the current state (relation), the candidates for
  the next gate and the gate which created this node (fgate). If you
  get the fgate fields from the first node to the last node, you obtain
  the network.
 ***********************************************************************/
typedef struct senode
{
  uint32_t ninputs;            // not really necessarily a field, but need
                          // to store it somewhere.
  uint32_t ngates;             // this is easy to discover, but it can be
                          // used for pruning etc. just for convenience
  rela_t *relation;
  i2tuple_t fgate;        // last gate formed to reach this node
  uint32_t *complexity;        // complexity of a line represents how many
                          // gates the line has gone thru. This can be
                          // used for ordering the candidate gates.
  uint32_t ncandi;
  i2tuple_t *candi;       // candidate gates from here on
  uint32_t upper_mask;    // upper_mask: mask of lines which occur in
                          // the upper inputs of the gates in this level
  uint32_t level_mask;    // level_mask: mask of lines which occur in 
                          // the gates in this level (both upper and lower
                          // inputs)
  sholist_t *kin;         // the remaining valid input bit strings
                          // after fgate is applied.
  struct senode *prev;
} senode_t;


#define SIGNTUPLE(a,b) ((a)*3+(b))

static candi_t *candi;
static int ncandi, scandi;

static void sorting_network_candidate(senode_t *prev, int a,int b);

static rela_t *gate_relation_transform(int N,rela_t *R, int a, int b)
{
  rela_t *C;
  int k;
  C= rela_dup(R);
  for(k=0;k<N;k++)
   if (k!=a && k!=b)
    switch(SIGNTUPLE(rela_compare(R,a,k), rela_compare(R,b,k)))
    {
    case SIGNTUPLE(-1,-1): 
      //  La < Lk and Lb < Lk => La' < Lk and Lb' < Lk
      //  There is no change
      break;
    case SIGNTUPLE(-1, 0):
      //  La < Lk and Lb ? Lk  => La' < Lk 
      //  There is no change
      break;
    case SIGNTUPLE(-1, 1):
      //  La < Lk and Lb > Lk 
      //  In this case,  La < Lk < Lb => La < Lb
      //  This can not be, the inputs of a gate must never be
      //  already related. The transition must be broken..
      //  So, this is also ignored.
      break;
    case SIGNTUPLE( 0,-1):
      //  La ? Lk  and Lb < Lk   =>  La' < Lk and Lb' ? Lk
      rela_less(C, a, k);
      rela_remove(C, b, k);
      break;
    case SIGNTUPLE( 0, 0): break;
    case SIGNTUPLE( 0, 1):
      //  La ? Lk  and Lb > Lk   =>  Lb' > Lk
      //  No change
      break;
    case SIGNTUPLE( 1,-1):
      //  La > Lk and Lb < Lk  =>  La > Lk > Lb => La > Lb
      //  This is again impossible. a and b must not be related
      //  Ignored.
      break;
    case SIGNTUPLE( 1, 0):
      //  La > Lk and Lb ? Lk => La' ? Lk  and Lb' > Lk
      rela_greater(C, b, k);
      rela_remove(C, a,k);
      break;
    case SIGNTUPLE( 1, 1): break;
      //  La > Lk and Lb > Lk => La' > Lk and Lb' > Lk
      //  no changes
      break;
    }
  rela_less(C, a,b);
  return C;
}

static sholist_t *apply_gate_to_sholist(sholist_t *L,int a,int b)
{
  sholist_t *C;
  uint16_t V;
  uint16_t mask, bad;
  int i;

  C= sholist_new(L->len);
  mask= (1<<a) | (1<<b);
  bad= (1<<a);
  for(i=0;i<L->len;i++)
  {
    V= L->data[i];
    if ((V&mask)==bad) V^= mask;
    C->data[i]= V;
  }
  unify_sholist(C);
  return C;
}

static void alloc_global_candi(int N)
{
  if (scandi < N*N)
  {
    if (candi) free(candi);
    scandi= N*N;
    candi= malloc(sizeof(candi[0])*scandi);
  }
}

static void candi_metric(senode_t *M, candi_t *C)
{
  int a,b;
  a= C->a;
  b= C->b;

  C->Mcom= imax2( M->complexity[a], M->complexity[b] );

  if (M->level_mask & ((1<<a) | (1<<b))) C->Mlevel= 1;
                                    else C->Mlevel= 0;

#if 0
  rela_t* RR;
  RR= gate_relation_transform(M->ninputs,M->relation, a, b);
  kdif= M->ninputs*M->ninputs/2 - rela_count0(M->relation) + rela_count0(RR);
  rela_free(RR);
  return (comp<<8) | dif | (kdif<<16) | (lev<<24);
#endif
}

static int candicompar(const void *_A, const void *_B)
{
  const candi_t *a= _A;
  const candi_t *b= _B;
  int R;
   // the sort is done in decreasing metric. this is
   // because I want to use the gates starting from the 
   // end to the start of the array.
  R= b->Mlevel - a->Mlevel;
  if (R) return R;
  R= b->Minset - a->Minset;
  if (R) return R;
  return b->Mcom - a->Mcom;
}

static void global_candi_to_node(senode_t *M)
{
  int i;
  for(i=0;i<ncandi;i++) candi_metric(M, candi+i);
  qsort(candi,ncandi,sizeof(candi[0]),candicompar);
  M->ncandi= ncandi;
  M->candi= malloc(sizeof(M->candi[0])*ncandi);
  for(i=0;i<ncandi;i++)
  {
    M->candi[i].a= candi[i].a;
    M->candi[i].b= candi[i].b;
  }
}

static gatelist_t *build_gatelist(senode_t *node, int fa, int fb)
{
  senode_t *K;
  gatelist_t *GL;
  int ng,p;

  GL= malloc(sizeof(GL[0]));
  GL->ninputs= node->ninputs;
  ng= GL->ngates= node->ngates+1;
  GL->gates= malloc(sizeof(GL[0])*ng);
  
  p= ng-1;
  GL->gates[p].A= fa; 
  GL->gates[p].B= fb;
  p--;
  for(K=node;K && p>=0;K=K->prev)
  {
    GL->gates[p].A= K->fgate.a;
    GL->gates[p].B= K->fgate.b;
    p--;
  }
  return GL;
}

static void senode_free(senode_t *M)
{
  rela_free(M->relation);
  free(M->complexity);
  if (M->kin) { free(M->kin->data); free(M->kin); }
  free(M);
}

void print_node(senode_t *M)
{
  int i;
  printf("NODE: %d inputs %d gates\n",M->ninputs, M->ngates);
  printf("Relation:\n");
  rela_print(M->relation);
  printf("forming gate: %d,%d\n",M->fgate.a,M->fgate.b);
  printf("candidates: ");
  for(i=0;i<M->ncandi;i++) printf("%d/%d ", M->candi[i].a,M->candi[i].b);
  printf("\n");
  printf("--------------END NODE ------------\n");
}

static int sholist_identical(sholist_t *A, sholist_t *B)
{
  if (A->len!=B->len) return 0;
  return !memcmp(A->data, B->data, A->len*sizeof(A->data[0]));
}

static void filter_candi_by_relation(rela_t *R)
{
  int i;
  for(i=0;i<ncandi;i++)
   if (rela_compare(R,candi[i].a,candi[i].b))
     { candi[i]= candi[ncandi-1]; i--; ncandi--; }
}

static void make_node_candidates(senode_t *M)
{
  int i,j,max_upper;
  int N;
  N= M->ninputs;

  alloc_global_candi(N);
  ncandi= 0;
  for(i=0;i<N;i++)
   for(j=i+1;j<N;j++)
     if (!rela_compare(M->relation, i,j))
      { candi[ncandi].a= i; candi[ncandi].b= j; ncandi++; }

#if 0
  /******** PRUNING FOR LEVEL PERMUTATIONS *********/
  for(i=N-1;i>=0;i--)
    if (M->upper_mask & (1<<i)) 
      break;
  max_upper= i;

  for(i=0;i<ncandi;i++)
         // if it forms a new level, we let it pass
    if ( ((1<<candi[i].a) | (1<<candi[i].b))&M->level_mask) continue;
         // it's in the same level, but in wrong order
    else if (candi[i].a<max_upper) { candi[i]= candi[ncandi-1]; ncandi--; i--; }
  /****************** END **************************/
#endif

  int rchange;
  rchange= 0;
  for(i=0;i<ncandi;i++)
  {
    sholist_t *Y;
    Y= apply_gate_to_sholist(M->kin, candi[i].a, candi[i].b);
    candi[i].Minset= Y->len;
    if (sholist_identical(M->kin, Y))
    {
      rela_less(M->relation, candi[i].a,candi[i].b);
      rchange= 1;
      // FIXME: should I also fix R here, by setting a < b?
      candi[i]= candi[ncandi-1]; ncandi--; i--;
    }
    free(Y->data); free(Y);
  }

  if (rchange) { // printf("updating relation due to input set\n");
    rela_apply_transitions(M->relation); 
    filter_candi_by_relation(M->relation); }

  global_candi_to_node(M);
}

senode_t *senode_make(senode_t *prev, int a, int b,int simple)
{
  senode_t *M;
  int N, sz;
  uint32_t gmask;
  rela_t *RR;

  N= prev->ninputs;

  if (prev->kin && prev->kin->len==N+1)
  {
    printf("via input set 2\n");
    sorting_network_candidate(prev->prev, prev->fgate.a, prev->fgate.b);
    return NULL;
  }

  RR= gate_relation_transform(N,prev->relation, a, b);

  if (rela_count0(RR)==0)
  {
    // we have reached a full relation, maybe this is a
    // sorting network?
    rela_free(RR);
    sorting_network_candidate(prev, a,b);
    return NULL;
  }

  M= calloc(1,sizeof(*M));
  M->ninputs= N;
  M->ngates= prev->ngates+1;
  M->relation= RR;
  M->fgate.a= a; 
  M->fgate.b= b;

  sz= sizeof(M->complexity[0])*N;
  M->complexity= malloc(sz);
  memcpy(M->complexity, prev->complexity, sz);
  M->complexity[a]= M->complexity[b]= 
                 imax2(prev->complexity[a],prev->complexity[b]) + 1;

  gmask= (1<<a) | (1<<b);

  if (gmask & prev->level_mask)
  {
    M->level_mask= gmask;
    M->upper_mask= 1<<a;
  }
  else
  {
    M->level_mask= prev->level_mask | gmask;
    M->upper_mask= prev->upper_mask | (1<<a);
  }

  M->prev= prev;

  if (simple) return M;

  M->kin= apply_gate_to_sholist(prev->kin, a,b);
#if 1
  if (M->kin->len==prev->kin->len)
  {
    // we should never reach here, actually.
    // this should be found when we're making the
    // candidate array.
    printf("useless move %d, %d, aborting\n", a,b);
    senode_free(M);
    return NULL;
  }
#endif

  make_node_candidates(M);
  return M;
}

senode_t *senode_init(int N)
{
  senode_t *M;
  int i,j;

  M= calloc(1,sizeof(*M));
  M->ninputs= N;
  M->relation= rela_new(N);
  M->complexity= calloc(N,sizeof(M->complexity[0]));
  
  alloc_global_candi(N);
  ncandi= 0;
  for(i=0;i<N;i++)
   for(j=i+1;j<N;j++)
     { candi[ncandi].a= i; candi[ncandi].b= j; ncandi++; }

  global_candi_to_node(M);
  return M;
}

static senode_t *top;
static network_t *best;

static int network_level(network_t *NW)
{
  return NW->gates[NW->ngates-1]->level;
}

static void sorting_network_candidate(senode_t *prev, int a,int b)
{
  gatelist_t *GL;
  network_t *NN;
  char *z;

  GL= build_gatelist(prev, a, b);
  NN= network_from_gate_list(GL, 0);
  z= gatelist_to_string(GL);
  gatelist_free(GL);

  printf("network: %s\n", z);
  free(z);

  if (!network_sorts(NN)) { printf("doesn't sort!\n"); goto worse; }

  network_levels(NN); 
  printf("sorts: ngates= %d nlevels= %d\n", NN->ngates, network_level(NN) + 1);

  if (!best) goto better;
  if (NN->ngates>best->ngates) goto worse;
  if (NN->ngates<best->ngates) goto better;
  if (network_level(NN)<network_level(best)) goto better;
worse:
  network_free(NN);
  return ;
better:
   printf("Best so far\n");
   best= NN;
   main_FOUND(best);
}

void relaser_process_node()
{
  senode_t *K;
  int a,b;
  int N;
again:
  if (!top) { printf("---- finished search!-----\n") ; 
              itf_pause_relaser(); return ; }

  N= top->ninputs;
  if (top->kin && top->kin->len==N+1)
  {
    printf("via input set\n");
    sorting_network_candidate(top->prev, top->fgate.a, top->fgate.b);
    goto free_node;
  }
  if ((best && top->ngates>=best->ngates-1) || !top->ncandi)
    goto free_node;
  a= top->candi[top->ncandi-1].a;
  b= top->candi[top->ncandi-1].b;
  top->ncandi--;
  K= senode_make(top, a,b, 0);
  if (K) top= K;
  return ;

free_node:
  K= top->prev;
  senode_free(top);
  top= K;
  goto again;
}


static void senode_write(FILE *f, senode_t *M)
{
  int i;
  if (!M) { write_int(f,0); return; }
  write_int(f,1);
  write_int(f, M->ninputs);
  write_int(f, M->ngates);
  write_int(f, M->fgate.a);
  write_int(f, M->fgate.b);
  write_int(f, M->upper_mask);
  write_int(f, M->level_mask);
  write_int(f, M->ncandi);
  for(i=0;i<M->ncandi;i++)
   { write_int(f, M->candi[i].a); write_int(f, M->candi[i].b); }
  intlist_write(f, M->ninputs, M->complexity);
  rela_write(f, M->relation);
  if (M->kin) { write_int(f, M->kin->len); 
                sholist_write(f, M->kin->len, M->kin->data); }
  else { write_int(f, 0); }
}

static senode_t *senode_read(FILE *f)
{
  uint32_t magic;
  uint32_t k;
  senode_t *M;
  int i;

  read_int(f,&magic);
  if (!magic) return NULL;

  M= calloc(1, sizeof(*M));

  read_int(f, &M->ninputs);
  read_int(f, &M->ngates);
  read_int(f, &M->fgate.a);
  read_int(f, &M->fgate.b);
  read_int(f, &M->upper_mask);
  read_int(f, &M->level_mask);
  read_int(f, &M->ncandi);
  M->candi= malloc(sizeof(M->candi[0])*M->ncandi);
  for(i=0;i<M->ncandi;i++)
   { read_int(f, &M->candi[i].a); read_int(f, &M->candi[i].b); }
  M->complexity= intlist_read(f, M->ninputs);
  M->relation= rela_read(f);
  read_int(f, &k);
  if (k) { M->kin= malloc(sizeof(M->kin[0]));
           M->kin->len= k;
           M->kin->data= sholist_read(f, M->kin->len); }
  return M;
}

void relaser_save(char *fn)
{
  FILE *f;
  senode_t *Z;
  gatelist_t *GL;

  f= fopen(fn,"wb");
  if (!f) 
  { 
   fprintf(stderr,"can't open %s for writing: %s\n", fn, strerror(errno));
   return ;
  }
  for(Z=top;Z;Z=Z->prev) senode_write(f, Z);
  senode_write(f,NULL);
  if (best)
  {
    uint32_t *K,i;
    write_int(f, 1);
    network_to_gate_list(best, &GL);
    K= malloc(sizeof(K[0])*GL->ngates*2);
    for(i=0;i<GL->ngates;i++) { K[2*i]= GL->gates[i].A;
                                K[2*i+1]= GL->gates[i].B; }
    write_int(f, GL->ngates);
    intlist_write(f, GL->ngates*2, K);
    gatelist_free(GL);
    free(K);
  }
  else write_int(f,0);
  fclose(f);
}


void relaser_load(char *fn)
{
  FILE *f;
  senode_t *Z;
  gatelist_t *GL;
  senode_t **target;
  uint32_t magic;

  best= NULL;
  top= NULL;
  target= &top;

  f= fopen(fn,"rb");
  if (!f) 
  { 
   fprintf(stderr,"can't open %s for reading: %s\n", fn, strerror(errno));
   return ;
  }
  while(1)
  {
    Z= senode_read(f);
    if (!Z) break;
    *target= Z;
    target= &Z->prev;
  }
  read_int(f,&magic);
  if (magic)
  {
    uint32_t ng, *K,i;
    read_int(f, &ng);
    K= intlist_read(f, ng*2);
    GL= malloc(sizeof(*GL));
    GL->ninputs= top->ninputs;
    GL->ngates= ng;
    GL->gates= malloc(sizeof(GL->gates[0])*ng);
    for(i=0;i<ng;i++)
     { GL->gates[i].A= K[2*i]; GL->gates[i].B= K[2*i+1]; }
    free(K);
    best= network_from_gate_list(GL, 0);
    gatelist_free(GL);
  }
  fclose(f);
  if (best) main_FOUND(best);
}



void relaser_run(int nnodes)
{
  itf_resume_relaser();
}

void relaser_pause()
{
  itf_pause_relaser();
  if (top) print_node(top);
}

void pbin(uint32_t V, int n)
{
  int i;
  for(i=n-1;i>=0;i--)
    if (V&(1<<i)) printf("1"); else printf("0");
}

void relaser_init(int N)
{
  int LL;
  int i,p;
  uint16_t V;

  best= NULL;
  top= senode_init(N);
  top->ncandi= 0;
  LL= 1;
  for(i=0;i<N/2;i++)
  {
    top= senode_make(top, 2*i, 2*i+1, 1);
    LL*= 3;
  }
  if (N%2) LL*= 2;
  printf("making input values past the wall\n");
  top->kin= sholist_new(LL);

  V= 0;
  p= 0;
                                    //  0  1  2   3
  static const unsigned char next[4]= { 2, 0, 3,  0 };
  int maxshif;

  maxshif= N; if (N%2) maxshif--;
  do {
    unsigned int shif, c;
    printf("init: "); pbin(V, N); printf("\n");
    top->kin->data[p++]= V;
    shif= 0;
    while(1) 
    {  
       c= (V >> shif)&3; c= next[c];
       V= (V & ~(3<<shif))  | (c<<shif);
       if (c!=0) break;
       shif+= 2; if (shif>=maxshif) break; 
    }
  } while(V!=0);

  if (N%2)
    for(i=0;i<p;i++) { top->kin->data[p+i]= top->kin->data[i] | (1<<(N-1)); }
  make_node_candidates(top);
  
  print_node(top);
}
