
#include <stdio.h>
#include "network.h"
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>


typedef struct mbox mbox_t;
typedef struct { int _msg; } msg_t;

static mbox_t *mbox_new();
static int mbox_fd(mbox_t *m);
static void mbox_put(mbox_t *m, void *data);
static void* mbox_get(mbox_t *m);

static void DISPATCH_main(msg_t *M);

#define _DEFM_QUIT 0
#define _DEFM_PAUSE 1
#define _DEFM_RESUME 2


struct mbox {
  pthread_mutex_t lock;
  int len, head,size;
  void **data;
  int fd[2];
};

typedef struct 
{
  mbox_t *mbox;
  void (*dispatch)(msg_t*);
  void (*workloop)();
  int type; // 0= normal 1= free running
  pthread_t systhread;
  int running; // not paused
} _thread_t;

static _thread_t THREAD_main;

static mbox_t *mbox_new()
{
   mbox_t *M;
   M= malloc(sizeof(*M));
   M->size= 64;
   M->data= malloc(M->size*sizeof(void*));
   M->head=0; M->len= 0;
   if (pthread_mutex_init(&M->lock, 0)) goto out_data;
   if (pipe(M->fd)) goto out_lock;
   return M;
out_lock:
   pthread_mutex_destroy(&M->lock);
out_data:
   free(M->data);
   free(M);
   return 0;
}

static int mbox_fd(mbox_t *m)
{
   return m->fd[0];
}

static void mbox_put(mbox_t *m, void *data)
{
   pthread_mutex_lock(&m->lock);
   if (m->len==m->size)
   {
      void **ND; int NS,i;
      NS= m->size*2;
      ND= malloc(sizeof(void*)*NS);
      for(i=0;i<m->len;i++)
        ND[i]= m->data[(m->head+i)%m->size];
      free(m->data);
      m->head= 0;
      m->size= NS;
      m->data= ND;
   }
   m->data[(m->head+m->len)%m->size]= data;
   m->len++;
   pthread_mutex_unlock(&m->lock); 
   write(m->fd[1], data, 1);
}

static void* mbox_get(mbox_t *m)
{
   char c;
   void *R;
again:
   read(m->fd[0], &c, 1);
   pthread_mutex_lock(&m->lock);
   if (m->len==0) {
      // the message was cleared, but we couldn't clear the byte in 
      // the pipe of course.
      pthread_mutex_unlock(&m->lock);
      goto again;
   }
   R= m->data[m->head];
   m->head= (m->head+1)%m->size;
   m->len--;
   pthread_mutex_unlock(&m->lock);
   return R;
}

static void LOOP_free_running(_thread_t *th)
{
  msg_t *M;
  int fd,R;
  fd_set rfd;
  struct timeval tm;

again:
  if (!th->running) goto do_message;
  fd= mbox_fd(th->mbox);
  FD_ZERO(&rfd);
  FD_SET(fd,&rfd);
  memset(&tm,0,sizeof(tm));
  R= select(fd+1, &rfd, NULL, NULL, &tm);
  if (R==0) { (*th->workloop)(); goto again; }
  if (R<0)  { if (errno==EINTR) goto again; else return ; }
do_message:
  M= mbox_get(th->mbox);
  switch(M->_msg)
  {
  case _DEFM_QUIT: free(M); return ;
  case _DEFM_RESUME: th->running= 1; free(M); break;
  case _DEFM_PAUSE: th->running= 0; free(M); break;
  default: (*th->dispatch)(M); break;
  }
  goto again;
}

static void* LOOP(void *_a)
{
  _thread_t *dd;
  msg_t *M;
  dd= _a;

  if (dd->type==1) 
  {
    LOOP_free_running(dd);
    return NULL;
  } 
  while(1)
  {
     M= mbox_get(dd->mbox);
     if (M->_msg==_DEFM_QUIT) { free(M); break; }
     (*dd->dispatch)(M);
  }
  return NULL;
}


int itf_init_main()
{
  THREAD_main.mbox= mbox_new();
  return mbox_fd(THREAD_main.mbox);
}

void itf_process_main()
{
  DISPATCH_main(mbox_get(THREAD_main.mbox));
}

int itf_pending_main()
{
  int fd,R;
  fd_set rfd;
  struct timeval tm;

again:
  fd= mbox_fd(THREAD_main.mbox);
  FD_ZERO(&rfd);
  FD_SET(fd,&rfd);
  memset(&tm,0,sizeof(tm));
  R= select(fd+1, &rfd, NULL, NULL, &tm);
  if (R==0) return 0;
  if (R<0)  { if (errno==EINTR) goto again; else return 0; }
  return 1;
}

static void thread_init(_thread_t *thr,void (*disp)(msg_t*),void (*worf)())
{
  pthread_attr_t attr;
  thr->mbox= mbox_new();
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  thr->dispatch= disp;
  thr->workloop= worf;
  if (worf) { thr->type=1; thr->running= 0; }
  pthread_create(&thr->systhread,&attr,&LOOP,thr);
}

static void _thread_pause(_thread_t *th)
{
  msg_t *M;
  M= malloc(sizeof(*M));
  M->_msg= _DEFM_PAUSE;
  mbox_put(th->mbox, M);
}

static void _thread_resume(_thread_t *th)
{
  msg_t *M;
  M= malloc(sizeof(*M));
  M->_msg= _DEFM_RESUME;
  mbox_put(th->mbox, M);
}




/*** IMPLEMENTATION ***/

void relaser_load(char *fn);
void relaser_save(char *fn);
void relaser_run(int nnodes);
void relaser_init(int ninputs);
void relaser_pause();
void main_found(network_t *nw);


/*** INTERFACE ***/

void relaser_LOAD(char *fn);
void relaser_SAVE(char *fn);
void relaser_RUN(int nnodes);
void relaser_INIT(int ninputs);
void relaser_PAUSE();
void main_FOUND(network_t *nw);
void itf_init_relaser();
void itf_quit_relaser();
void itf_pause_relaser();
void itf_resume_relaser();
void relaser_process_node();
int itf_init_main();
void itf_process_main();
int itf_pending_main();



static _thread_t THREAD_relaser;



typedef struct {
  int _msg;
  char *fn;
} relaser_LOAD_t;
typedef struct {
  int _msg;
  char *fn;
} relaser_SAVE_t;
typedef struct {
  int _msg;
  int nnodes;
} relaser_RUN_t;
typedef struct {
  int _msg;
  int ninputs;
} relaser_INIT_t;
typedef struct {
  int _msg;
} relaser_PAUSE_t;
typedef struct {
  int _msg;
  network_t *nw;
} main_FOUND_t;



static void DISPATCH_relaser(msg_t *_M)
{
  switch(_M->_msg) {
    case 3: { 
     relaser_LOAD_t *_V= (void*) _M;
     relaser_load(_V->fn);
 free(_V);
    } break;
    case 4: { 
     relaser_SAVE_t *_V= (void*) _M;
     relaser_save(_V->fn);
 free(_V);
    } break;
    case 5: { 
     relaser_RUN_t *_V= (void*) _M;
     relaser_run(_V->nnodes);
 free(_V);
    } break;
    case 6: { 
     relaser_INIT_t *_V= (void*) _M;
     relaser_init(_V->ninputs);
 free(_V);
    } break;
    case 7: { 
     relaser_PAUSE_t *_V= (void*) _M;
     relaser_pause();
 free(_V);
    } break;
  }
}
static void DISPATCH_main(msg_t *_M)
{
  switch(_M->_msg) {
    case 3: { 
     main_FOUND_t *_V= (void*) _M;
     main_found(_V->nw);
 free(_V);
    } break;
  }
}



void relaser_LOAD(char *fn)
{
  relaser_LOAD_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 3;
  _V->fn= fn;
  mbox_put(THREAD_relaser.mbox,_V);
}
void relaser_SAVE(char *fn)
{
  relaser_SAVE_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 4;
  _V->fn= fn;
  mbox_put(THREAD_relaser.mbox,_V);
}
void relaser_RUN(int nnodes)
{
  relaser_RUN_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 5;
  _V->nnodes= nnodes;
  mbox_put(THREAD_relaser.mbox,_V);
}
void relaser_INIT(int ninputs)
{
  relaser_INIT_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 6;
  _V->ninputs= ninputs;
  mbox_put(THREAD_relaser.mbox,_V);
}
void relaser_PAUSE()
{
  relaser_PAUSE_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 7;
  mbox_put(THREAD_relaser.mbox,_V);
}
void main_FOUND(network_t *nw)
{
  main_FOUND_t*_V;
  _V= malloc(sizeof(*_V));
  _V->_msg= 3;
  _V->nw= nw;
  mbox_put(THREAD_main.mbox,_V);
}



void itf_init_relaser(){
  thread_init(&THREAD_relaser,DISPATCH_relaser,relaser_process_node);
}
void itf_quit_relaser(){
  msg_t *M;
  void *ret;
  M= malloc(sizeof(*M));
  M->_msg= 0;
  mbox_put(THREAD_relaser.mbox,M);
  pthread_join(THREAD_relaser.systhread,&ret);
}
void itf_pause_relaser(){ _thread_pause(&THREAD_relaser); }
void itf_resume_relaser(){ _thread_resume(&THREAD_relaser); }


